import osmnx as ox
import networkx as nx
import folium
import itertools
import pickle
import time, datetime
import os.path
import sys
import webbrowser

import random
 
import os
import glob


from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.ops import linemerge

import matplotlib.pyplot as plt
plt.switch_backend('agg')


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)
print('Python version: ',sys.version)

start_time = time.time()

water_tags = {'natural':['spring','water'],'water':['pond','lake'],
              'landuse':['reservoir','basin'],'waterway':['river','stream']
        }


# fix later, convert features to waypoints (nodes) somehow
bbox_gila = (34.35704, 32.77935, -107.65983, -108.94574) # rough box of Gila NF
wgdf = ox.features.features_from_bbox(bbox=bbox_gila, tags=water_tags)
points = wgdf.copy()
points['geometry'] = points['geometry'].centroid
print(points.head())
#points.to_file('gxl_water', driver='GeoJSON')
WG = ox.utils_graph.graph_from_gdfs(points, None, graph_attrs=None)
ox.io.save_graph_geopackage(WG, filepath= 'gxl_water + '.gpkg)


##wgdf.to_file('gxl_water.gpx', driver='GPX')
##
def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)

def graph_length(Q):
    length = ox.stats.edge_length_total(Q) #in meters
    length = length * 0.0006213712
    return length

def get_bounds(Q):
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    tb = edges.total_bounds
    print( tb )
    sw = [tb[1],tb[0]]
    ne = [tb[3],tb[2]]
    return [sw, ne]

def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    fours = 0
    four_pairs = 0
    parallels = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
        if d==4:
            fours += 1
            for neighbor in Q.neighbors(node):
                if Q.degree(neighbor)==4 and neighbor>node: # prevent double-count
                    four_pairs += 1
        if d>6:
            print('node ', node, 'has degree ', d)
        for neighbor in Q.neighbors(node):
            edge_count = Q.number_of_edges(node,neighbor)
            if edge_count > 1:
                parallels += edge_count - 1
                
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of 4-way intersections ' + txt + ' :', fours)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('number of parallel edges ' + txt + ' :', parallels)
    if four_pairs > 0:
        print('number of 4-way adjacent pairs ' + txt + ' :', four_pairs)

        
    check_node_geometry(Q)
    print('')

def my_cmap(value):
    return value

def colorize_graph(Q):
    QR = Q.copy()
    for node, dat in Q.nodes(data=True):
        d = Q.degree(node)
        dat['color'] = d

        QR.add_node(node,**dat)                #replace node value

    for u, v, k, dat in Q.edges(keys=True,data=True):
        d = 'path'
        if 'highway' in dat:
            d = dat['highway']
        dat['color']=0.0
        if isinstance(d, str):
            if d=='path' or d == 'footway':
                dat['color']=1.0
            else:
                dat['color'] = 2.0
        QR.add_edge(u,v,key=k,**dat)
    QR.graph['color']=True
    return QR

def colorize_self_loops(Q,color=9.0):
    QR = colorize_graph(Q)
    for u,v,k,dat in QR.edges(keys=True,data=True):
        if u==v:
            QR[u][v][k]['color'] = color
            #dat['color']=color
            #QR.add_edge(u,v,key=k,**dat)
    return QR
    
                 
def colorize_edge_list(Q,edge_list,color=10.0):
     QR = colorize_graph(Q)
     for u,v in edge_list:
         dat = QR.nodes[u]
         dat['color'] = color
         QR.add_node(u,**dat)
         dat = QR.nodes[v]
         dat['color'] = color
         QR.add_node(v,**dat)
         if QR.has_edge(u,v):
             for k in QR[u][v]:
                 QR[u][v][k]['color']= color 
            # for v1,k in QR[u]:
             #    if v1==v:
              #       QR.edges[u,v,k]['color'] = color  # fix key
        #else if QR.has_edge(v,u):
        #    QR.edges[v,u,0]['color'] = 10.0
     return QR

def colorize_parallel_edges(Q):
    return colorize_edge_list(Q,parallel_edges(Q),color=9.0)

def parallel_edges(Q):
    ret_edges = set()
    for u,v in Q.edges():
        edge_count = Q.number_of_edges(u,v)
        if edge_count>1:
            ret_edges.add( ( min(u,v),max(u,v) ) )
    print('number of parallel edges: ',len(ret_edges))
    return ret_edges
    
def colorize_mismatched_nodes(Q,node_pairs,color=10.0):
    QR = colorize_graph(Q)
    matched_nodes = set()
    unmatched_nodes = set()
    for u,v in node_pairs:
        matched_nodes.add(u)
        matched_nodes.add(v)
    for node, dat in Q.nodes(data=True):
        if node not in matched_nodes:
            unmatched_nodes.add(node)
            dat['color'] = color
            QR.add_node(node,**dat) #update node attributes
            print('node ', node, ' not in matching')
    check_remainder_odd_nodes(J,unmatched_nodes)
    return QR

def check_node_geometry(Q):
    count = 0
    for node, dat in Q.nodes(data=True):
        if 'x' not in dat:
            count += 1
            #d = Q.degree(node)
            #print('node ', node, ' has no geometry attribute, degree ', d)t
    if count>0:
        print('number of nodes without geometry: ',count)


def check(Q,Q_original):
    remainder_nodes = get_remainder_odd_nodes(Q)
    #print('number of odd nodes remaining: ', len(remainder_nodes))
    check_remainder_odd_nodes(Q_original,remainder_nodes)
    for u,v,k,dic in Q.edges(data=True, keys=True):
        if 'node4_path' in dic:
            print('temporary path not deleted between: ',u,' and ',v, ' with key ', k)
        if not 'highway' in dic:
            print('suspicious edge: ',u,' and ',v, ' with key ', k)


def get_remainder_odd_nodes(Q):
    node_list = []
    for node in Q.nodes:
        if Q.degree(node) %2 ==1:
            node_list.append(node)
    return node_list

# remove trailheads and termini
def delete_ends(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            if len(neighbors) == 1:
                rn.append(node)
        if len(rn)>0:            
            Q.remove_nodes_from(rn)
        else:
            break

def check_remainder_odd_nodes(Q_original,node_list):
    remainder_odd_node_count = 0
    remainder_odd_node_4node_neighber_count = 0
    for node in node_list:
        if Q_original.degree(node)==4:
            remainder_odd_node_count +=1
            #print('remainder odd node ', node, 'was a 4node')
        else:
            for neighbor in Q_original.neighbors(node):
                if Q_original.degree(neighbor)==4:
                    #print('remainder odd node ',node,' has neighbor ', neighbor,' that is a 4node')
                    remainder_odd_node_4node_neighber_count +=1
    if remainder_odd_node_count > 0:
        print('remainder odd_node count: ', remainder_odd_node_count)
    if remainder_odd_node_4node_neighber_count > 0:
        print('remainder odd node with 4node neighbor count: ', remainder_odd_node_4node_neighber_count)
        

def check_self_loops(Q,txt=''):
    self_loops = 0
    self_loop_distance = 0
    max_self_loop_distance = 0
    for u,v,k,dic in Q.edges(keys=True,data=True):
        if u==v:
            d = dic['length']
            self_loops += 1
            self_loop_distance += d
            if d>max_self_loop_distance:
                max_self_loop_distance = d
    print('number of self-loops ' + txt + ' : ', self_loops )
    print(f'total self-loop distance {self_loop_distance*0.0006213712:.2f} miles ', txt)    
    print(f'maximum self-loop distance {max_self_loop_distance*0.0006213712:.2f} miles ', txt)
    print('')

def delete_self_loops(Q):
    self_loops = []
    for u,v,k,dic in Q.edges(keys=True,data=True):
        if u==v:
            self_loops.append( (u,v,k) )
    Q.remove_edges_from( self_loops)
    delete_ends(Q) # side-effect is that we left dangling ends
    
    
    
     
town_markers = [(34.29963, -108.13268, 'Pie Town'),
         (33.19593, -108.20834, 'Gila Hot Springs'),
         (32.77981, -108.27533, 'Silver City'),
         (33.37939, -108.90369, 'Alma/Glenwood'),
         (33.03022, -108.16864, 'Lake Roberts'),
         (32.85594, -107.98024, 'Mimbres'),       
         (32.91685, -107.70593, 'Kingston')]

# draw() changed after moving to Linux
def draw(Q,title='',show_nodes = False, boxes=None, not_boxes = None, markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
    T = Q.copy()
    if not 'color' in T.graph:
        T = colorize_graph(T)
    QD = T.to_directed()
    QD.graph['crs'] = ox.settings.default_crs

    # our graphs may be HUGE, so simplify if not already done so
    if not "simplified" in QD.graph:
        QD = ox.simplify_graph(QD, endpoint_attrs=None, remove_rings=True, track_merged=True)
    
    nodes, edges = ox.graph_to_gdfs(QD)
#    m = edges.explore(color='#ff00ff', tiles="openstreetmap",tooltip='name')
    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap",tooltip='name')
    if show_nodes:    
#        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    
        nodes.explore(m=m, column='color', cmap='tab10', legend=False,marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        kw = {"prefix": "fa", "color": "green", "icon": "arrow-up"}
        icon = folium.Icon(angle=0, **kw)
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          icon=folium.Icon(color='lightred',icon=''),
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
   # folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
    #m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)    
        
def output_gpx(Q, title='foo',markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    T = Q.copy()
    QD = T.to_directed()
    QD.graph['crs'] = ox.settings.default_crs

    # our graphs may be HUGE, so simplify if not already done so
#    if not "simplified" in QD.graph:
#        QD = ox.simplify_graph(QD, endpoint_attrs=None, remove_rings=True, track_merged=True)

    ox.io.save_graph_geopackage(Q, filepath= title + '.gpkg')
    
    nodes, edges = ox.graph_to_gdfs(QD)
    
    # credit https://pygis.io/docs/d_access_osm.html
  #  edges  = edges.loc[:,edges.columns.str.contains('addr:|geometry')]
  #  edges = edges.loc[edges.geometry.type=='LineString']
  #  edges.to_file(title + '_edges.gpx', driver='GPX')

    
    #edges.to_file(title + '.shp')
   # nodes.to_file(title + '_nodes.shp')
    #todo also add markers
    #todo also import features such as water

         


K = pickle.load(open('max_candidate_with_towns.pkl', 'rb'))
#report(K,'candidate graph ')

       
if nx.is_eulerian(K):
    print('k is eulerian!!!!!')
else:
    print('back to the drawing board')
    K = colorize_graph(K)
    draw(K,'back to drawing board',show_nodes=True)

output_gpx(K,'gxl')

            
               

def cleanup_graph(Q):
    QR = nx.MultiGraph()
    bad_nodes = [240288183,2480249190]
    for u,v,k,d in Q.edges(keys=True, data=True):
        if u in bad_nodes or v in bad_nodes:
            continue
        QR.add_edge(u,v,k,**d)
    for node, d in Q.nodes(data=True):
        if node in bad_nodes:
            continue
        QR.add_node(node,**d)   #replace node value
    QR.graph['crs'] = ox.settings.default_crs    
    return QR
        

def fix_node_geometry(Q,Q_original):
    restore_count = 0
    QR = Q.copy()
    for node in Q.nodes():  # add node attributes back in
        node_dic = Q.nodes[node]
        if 'x' not in node_dic:
            node_dic = Q_original.nodes[node]
            QR.add_node(node,**node_dic)
            restore_count += 1
   # if restore_count > 0:
    #    print('number of 4nodes restored: ', restore_count)
    return QR
    
#K = fix_node_geometry(K,J)       
    


#draw(K,'max_candidate_with_towns')


elapsed_time(start_time)

print('finished script ' + os.path.basename(__file__))






